#
# CVSFile and CVSDirectory
#
# Files and directories managed with CVS
#
# Copyright (c) 2001-2003 by Philipp "philiKON" von Weitershausen
#
# This software is distributed under the terms of the Mozilla Public License
# (MPL)
#

"""
$Id: CVSObjects.py,v 1.2 2003/03/17 13:49:49 philipp Exp $
"""

import os
from Persistence import Persistent

from interfaces.RevisionControlled import IRCFile, IRCDirectory

# default status
default_status = {'filename': '',
                  'rpath':    '',
                  'status':   '',
                  'wrev':     '',
                  'rrev':     '',
                  'stag':     '',
                  'sdate':    '',
                  'sopt':     '',
                  'tags':     {} }

def posix_path_join(*args):
    if not len(args):
        return ""
    if not args[0]:
        args = args[1:]
    return "/".join(args)

class CVSFile(Persistent):
    """
    A file managed with CVS
    """

    __implements__ = IRCFile

    __allow_access_to_unprotected_subobjects__ = 1

    isDirectory = 0

    def __init__(self, module_path, relative_path, status=None):
        # attributes defined in interface:
        self.filename = os.path.basename(relative_path)
        self.relative_path = relative_path
        self.absolute_path = os.path.join(module_path, relative_path)
        
        self._module_path = module_path
        self._status = {}
        
        if status is None:
            status = default_status.copy()
        statusdict = { self.relative_path: default_status.copy() }
        self.setStatusInfo(statusdict)

    def getStatus(self):
        return self._status.copy()

    def getStatusInfo(self, status=None):
        if status is None:
            # always make a unique dict
            status = {}
        status[self.relative_path] = self._status
        return status

    def setStatusInfo(self, statusdict):
        if not statusdict.has_key(self.relative_path):
            return
        # it is important that we use the update() method here because the
        # statusdict might only contain updates, not a complete set of status
        # information
        self._status.update(statusdict[self.relative_path])
        self._p_changed = 1

class CVSDirectory(Persistent):
    """
    A directory containing files managed with CVS
    """

    __implements__ = IRCDirectory

    __allow_access_to_unprotected_subobjects__ = 1

    isDirectory = 1

    def __init__(self, module_path, relative_path, status=None):
        # attributes defined in interface:
        self.relative_path = relative_path
        if relative_path == "":
            self.filename = os.path.basename(module_path)
            self.absolute_path = module_path
        else:
            self.filename = os.path.basename(relative_path)
            self.absolute_path = os.path.join(module_path, relative_path)

        self._module_path = module_path
        self._contents = {}
        
        if status is not None:
            self.setStatusInfo(status)

    def getStatusInfo(self, status=None, recursive=1):
        if status is None:
            status = {}
        for obj in self.getContents():
            if not obj.isDirectory:
                status = obj.getStatusInfo(status)
            elif recursive:
                # directory, but only in recursive mode
                status = obj.getStatusInfo(status, recursive)
        return status

    def setStatusInfo(self, statusdict):
        # simply pass along the information to subobjects
        for obj in self.getContents():
            obj.setStatusInfo(statusdict)

    def hasFile(self, filename):
        return filename in self.getDirectoryListing()
        
    def getFileObject(self, filename):
        if not self.hasFile(filename):
            raise ValueError, "File does not exist: %s" % filename
        
        if self._contents.has_key(filename):
            return self._contents[filename]
        
        abs_path = os.path.join(self.absolute_path, filename)
        rel_path = posix_path_join(self.relative_path, filename)

        if os.path.isfile(abs_path):
            obj = CVSFile(self._module_path, rel_path)
        elif os.path.isdir(abs_path) and filename != "CVS":
            obj = CVSDirectory(self._module_path, rel_path)
        else:
            return None

        self._contents[filename] = obj
        self._p_changed = 1
        return obj
    
    def getContents(self):
        objects = []
        contents = {}
        for filename in self.getDirectoryListing():
            obj = self.getFileObject(filename)
            objects.append(obj)
            contents[filename] = obj
        self._contents = contents
        return objects

    def getDirectoryListing(self, recursive=0):
        if not os.path.isdir(self.absolute_path):
            return []
        listing = os.listdir(self.absolute_path)
        # filter dotted files
        listing = filter(lambda path:not path.startswith('.'), listing)
        # do not show the 'CVS' management directory
        if "CVS" in listing:
            listing.remove("CVS")
        listing.sort()
        if recursive:
            for obj in self.getContents():
                if obj.isDirectory:
                    for filename in obj.getDirectoryListing(recursive=1):
                        listing.append(posix_path_join(obj.filename, filename))
        return listing

    def traverse(self, path, default=None):
        if path == self.relative_path:
            # it's me!
            return self
        elif self.hasFile(path):
            # I have it
            return self.getFileObject(path)
        else:
            segments = path.split('/')
            if not (segments and self.hasFile(segments[0])
                    and self.getFileObject(segments[0]).isDirectory):
                return default
            next = segments[0]
            segments = segments[1:]
            return self.getFileObject(next).traverse(posix_path_join(*segments))
