#
# RevisionManager
#
# Copyright (c) 2001 by beehive elektronische medien GmbH
# Copyright (c) 2001-2003 by Philipp "philiKON" von Weitershausen
#
# This software is distributed under the terms of the Mozilla Public License
# (MPL)
#

"""
$Id: RevisionManager.py,v 1.34 2003/03/17 13:49:49 philipp Exp $
"""

# standard python modules
import os
from StringIO import StringIO

# Zope modules
import Globals
from OFS import SimpleItem, Image
from AccessControl import Role, ClassSecurityInfo
from App.Dialogs import MessageDialog
from Acquisition import aq_parent, aq_base, aq_acquire
from Products.PageTemplates.PageTemplateFile import PageTemplateFile

from ZPublisher.HTTPRequest import HTTPRequest
from ZPublisher.HTTPResponse import HTTPResponse
from ZPublisher.mapply import mapply

# product modules
import Permissions
from CVSWrapper import CVSWrapper, CVSError
from CVSObjects import CVSDirectory, posix_path_join
from FactoryRegistry import FactoryRegistry

from interfaces.RevisionManager import IRevisionManager, IRevisionManagerZMI

# constants and global variables
FS_DIRNAME = "RevisionManager"
DATA_DIR = os.path.join(Globals.INSTANCE_HOME, "var", FS_DIRNAME)
CVSPASSFILE = os.path.join(DATA_DIR, 'cvspass')
BAD_IDS = ('acl_users', 'RevisionManager', 'Control_Panel')

class TreeObjectWrapper:
    """
    A wrapper around Zope objects to be more flexible when displaying them
    with ZopeTree.
    """

    __allow_access_to_unprotected_subobjects__ = 1

    def __init__(self, path, obj, module_paths=[]):
        self.path = path
        self.object = obj
        self.id = obj.getId()
        self.folderish = obj.isPrincipiaFolderish
        self._module_paths = module_paths

    def getChildren(self):
        if not self.folderish:
            return []
        children = []
        objectIds = self.object.objectIds()
        objectIds.sort()
        for id in objectIds:
            # filter for bad IDs
            if id in BAD_IDS:
                continue
            path = posix_path_join(self.path, id)
            obj = getattr(self.object, id)
            children.append(TreeObjectWrapper(path, obj, self._module_paths))
        return children

    def supportsSourceMethods(self, wrapped_obj):
        """
        Check whether an object supports FTP/DAV protocols
        """
        if wrapped_obj.folderish:
            return 1
        base = aq_base(wrapped_obj.object)
        # object needs to support at least one method to retrieve its raw source
        # and one method to upload text to it
        return ((hasattr(base, 'manage_FTPget') or hasattr(base, 'EditableBody')
              or hasattr(base, 'document_src')  or hasattr(base, 'read'))
            and (hasattr(base, 'manage_upload') or hasattr(base, 'PUT')))

    def isNotInModule(self, wrapped_obj):
        """
        Check whether the object that is wrapped is available as a file in the module
        """
        if wrapped_obj.folderish and wrapped_obj.path in self._module_paths:
            # even if a folder is already in the module, one of its children
            # could still not be in
            return len(wrapped_obj.getChildrenNotInModule()) > 0
        return wrapped_obj.path not in self._module_paths

    def getChildrenSupportingSourceMethods(self):
        children = self.getChildren()
        return filter(self.supportsSourceMethods, children)

    def getChildrenInModule(self):
        children = self.getChildren()
        return filter(lambda child, paths=self._module_paths: child.path in paths, children)

    def getChildrenNotInModule(self):
        children = self.getChildren()
        return filter(self.isNotInModule, children)

class RevisionManager(SimpleItem.SimpleItem):
    """
    RevisionManager class
    """

    __implements__ = IRevisionManager, IRevisionManagerZMI

    # Zope specific stuff
    manage_options = (
        ( {'label': 'Status', 'action': 'rm_statusForm',
           'help': ('RevisionManager', 'status.stx') },
          {'label': 'Add objects', 'action': 'rm_addObjectsForm',
           'help': ('RevisionManager', 'addObjects.stx') },
          {'label': 'Map files to objects', 'action': 'rm_mapFSObjectsForm',
           'help': ('RevisionManager', 'mapFSObjects.stx') },
          {'label': 'Configuration', 'action': 'rm_configForm',
           'help': ('RevisionManager', 'config.stx')},
        )
        + SimpleItem.Item.manage_options
        + Role.RoleManager.manage_options
   )

    id = "RevisionManager"
    meta_type = "Revision Manager"
    security = ClassSecurityInfo()

    def __init__(self):
        # this title is temporary and will be replaced with the proper one
        # later on
        self.title = "not properly initialized"

        # private attributes
        self._settings = {}
        # create the data directory if it doesn't exist already
        if not os.path.isdir(DATA_DIR):
            os.mkdir(DATA_DIR)

    def _setId(self, id):
        # Don't permit rename
        if id != self.id:
            return MessageDialog(title="Invalid ID",
                                 message="It is not possible to change the ID of a RevisionManager object.",
                                 action="manage_workspace")

    def _canCopy(self, op=0):
        # Don't permit copying/moving
        return 0

    # private auxiliary methods:

    def _clean_up(self):
        settings = self._settings
        if not settings.has_key('module'):
            return
	module_path = os.path.join(settings['work_dir'], settings['module'])
	if os.path.isdir(module_path):
	    try:
	        deltree(module_path)
	    except OSError:
		pass

    def _create_wrapper(self):
        settings = self._settings
        self._cvs_wrapper = CVSWrapper(**settings)

    def _create_module_object(self):
        settings = self._settings
        if not settings.has_key('work_dir'):
            # work-around for update from 1.0 to 1.1
            work_dir = settings['work_dir'] = DATA_DIR
        else:
            work_dir = settings['work_dir']
        module = settings['module']
        module_path = posix_path_join(work_dir, module)
        self.module = CVSDirectory(module_path, "")

    def _create_fake_request(self):
        environ = os.environ.copy()
        environ['SERVER_NAME'] = 'localhost'
        environ['SERVER_PORT'] = ''
        response = HTTPResponse(stdout=StringIO())
        request = HTTPRequest(StringIO(''), environ, response)
        return request

    def _get_source(self, obj, request):
        """
        Get the source of an object
        """
        if (hasattr(aq_base(obj), 'data') and hasattr(obj.data, '__class__')
            and obj.data.__class__ is Image.Pdata):
            # yuck! we have a File instance with chunked data, let's stream it!
            result = ""
            data = obj.data
            while data is not None:
                result += data.data
                data = data.next
            return result

        kw = { 'REQUEST': request, 'RESPONSE': request.RESPONSE }
        
        if hasattr(obj, 'document_src'):
            result = mapply(obj.document_src, keyword=kw)
        elif hasattr(obj, 'read'):
            result = obj.read()
        elif hasattr(obj, 'EditableBody'):
            result = obj.EditableBody()
        elif hasattr(obj, 'manage_FTPget'):
            result = mapply(obj.manage_FTPget, keyword=kw)
        else:
            result = ""

        return str(result)

    def _put_source(self, obj, src):
        """
        Upload text to an object
        """
        request = self._create_fake_request()
        request.other['BODY'] = src
        base = aq_base(obj)
        if hasattr(base, "manage_upload"):
            obj.manage_upload(src)
        elif hasattr(base, "PUT"):
            kw = { 'REQUEST': request, 'RESPONSE': request.RESPONSE }
            mapply(obj.PUT, keyword=kw)
        else:
            return 1  # error
        return 0

    def _escape_spaces(self, paths):
        return [path.replace(' ', '\ ') for path in paths]

    def goToSecondStage(self, REQUEST=None):
        try:
            self.checkout()
        except CVSError:
            pass
        
        if self.haveWorkingCopy():
            # we're properly initialized
            self.title = self.id
            method = 'rm_mapFSObjectsForm'
        else:
            method = 'rm_importForm'

        if REQUEST is not None:
            REQUEST.RESPONSE.redirect(self.absolute_url() + '/' + method)

    ############################################################################
    # Implementation methods for interface
    # interfaces.RevisionManager.IRevisionManager
    #

    # settings

    security.declarePublic("getSettings")
    def getSettings(self):
        return self._settings.copy()

    security.declareProtected(Permissions.RMManage, "setSettings")
    def setSettings(self, settings):
        self._settings.update(settings)
        self._p_changed = 1

    security.declareProtected(Permissions.RMManage, "saveSettings")
    def saveSettings(self, mapping, module=None):
        settings = {
            'cvs_bin':         mapping.get("cvs_bin",         "cvs"),
            'access_method':   mapping.get("access_method",   "local"),
            'repository_path': mapping.get("repository_path", "/var/cvs"),
            'work_dir':        DATA_DIR,
            'server_addr':     mapping.get("server_addr",     ""),
            'user_name':       mapping.get("user_name",       ""),
            'passwd':          mapping.get("passwd",          ""),
            'cvs_rsh':         mapping.get("cvs_rsh",         "rsh"),
            'comp_level':      mapping.get("comp_level",      "3"),
            'cvs_passfile':    CVSPASSFILE,
            'debug':           mapping.get("debug",           0)
            }
        if module is not None:
            settings['module'] = module
        self.setSettings(settings)
        self._create_wrapper()
        if module is not None:
            self._create_module_object()

    # lazy traverse

    security.declarePrivate("lazyTraverse")
    def lazyTraverse(self, path, default=None):
        # if lazy traverse has not been initialized, work like
        # regular restrictedTraverse()
        if not hasattr(self, '_v_lazy_traverse'):
            self._v_lazy_traverse = {}
        
        if not self._v_lazy_traverse.has_key(path):
            if path == "":
                # this is a special behaviour
                # an empty path means the root directory of the module
                obj = self.getModuleFolder()
            else:
                obj = self.restrictedTraverse(path, default)
            # set the shortcut
            self._v_lazy_traverse[path] = obj
            return obj

        # the shortcut is present, so use it
        return self._v_lazy_traverse[path]

    security.declarePrivate("setLazyTraverse")
    def setLazyTraverse(self, path, obj):
        # do nothing if lazy traverse has not been initialized
        if not hasattr(self, '_v_lazy_traverse'):
            self._v_lazy_traverse = {}
        self._v_lazy_traverse[path] = obj

    security.declarePrivate("clearLazyTraverse")
    def clearLazyTraverse(self):
        if hasattr(self, '_v_lazy_traverse'):
            del self._v_lazy_traverse

    # 'getters'

    security.declareProtected(Permissions.RMManage, "getModuleFolder")
    def getModuleFolder(self):
        return aq_parent(self)

    security.declareProtected(Permissions.RMManage, "getTreeWrappedModuleFolder")
    def getTreeWrappedModuleFolder(self):
        module_folder = self.getModuleFolder()
        module_paths = self.getAllFilePaths()
        return TreeObjectWrapper("", module_folder, module_paths)

    security.declareProtected(Permissions.RMManage, "getCVSModule")
    def getCVSModule(self):
        if not hasattr(self, 'module'):
            self._create_module_object()
        return self.module

    security.declareProtected(Permissions.RMManage, "traverseCVSObject")
    def traverseCVSObject(self, path):
        module = self.getCVSModule()
        return module.traverse(path)

    security.declareProtected(Permissions.RMManage, "getAbsoluteFilePath")
    def getAbsoluteFilePath(self, path):
        settings = self._settings
        return os.path.join(settings['work_dir'], settings['module'], path)

    security.declareProtected(Permissions.RMManage, "getSupportedMetaTypes")
    def getSupportedFactoryMetaTypes(self):
        meta_types = FactoryRegistry.getSupportedMetaTypes()
        meta_types = list(meta_types)
        meta_types.sort()
        return tuple(meta_types)

    def getSubobjectsPaths(self, path, obj=None):
        if obj is None:
            obj = self.lazyTraverse(path)
        if obj is None:
            # someone has given us a fake path
            return []
        if not obj.isPrincipiaFolderish:
            return []
        paths = []
        for id, subobj in obj.objectItems():
            # filter for bad IDS
            if id in BAD_IDS:
                continue
            subpath = posix_path_join(path, id) 
            paths.append(subpath)
            # while we're at it, remember the object for lazy traversing
            self.setLazyTraverse(subpath, subobj)
            if obj.isPrincipiaFolderish:
                paths += self.getSubobjectsPaths(subpath, subobj)
        return paths

    security.declareProtected(Permissions.RMManage, "getParentPaths")
    def getParentPaths(self, path):
        parent_paths = []
        segments = path.split("/")
        for i in range(1, len(segments)):
            dir_path = "/".join(segments[:i])
            parent_paths.append(dir_path)
        return parent_paths

    security.declareProtected(Permissions.RMManage, "getAllFilePaths")
    def getAllFilePaths(self):
        module = self.getCVSModule()
        return module.getDirectoryListing(recursive=1)

    security.declareProtected(Permissions.RMManage, "getParentAndChildObjectPaths")
    def getParentAndChildObjectPaths(self, paths):
        all_paths = paths[:]
        for path in paths:
            # add parent directory path segments
            for parent_path in self.getParentPaths(path):
                if parent_path not in all_paths:
                    all_paths.append(parent_path)
            # add sub-objects paths
            all_paths += self.getSubobjectsPaths(path)
        all_paths.sort()
        return all_paths

    security.declareProtected(Permissions.RMManage, "getParentAndChildFilePaths")
    def getParentAndChildFilePaths(self, paths):
        all_paths = paths[:]
        for path in paths:
            # add parent directory path segments
            for parent_path in self.getParentPaths(path):
                if parent_path not in all_paths:
                    all_paths.append(parent_path)
            # add directory contents
            cvsobj = self.traverseCVSObject(path)
            if cvsobj.isDirectory:
                for subpath in cvsobj.getDirectoryListing(1):
                    all_paths.append(posix_path_join(path, subpath))
        all_paths.sort()
        return all_paths

    security.declareProtected(Permissions.RMManage, "getFSFilesNotInZODB")
    def getFSFilesNotInZODB(self):
        module_path = self.getCVSModule().absolute_path
        all_fspaths = self.getAllFilePaths()
        notinzodb = []
        for path in all_fspaths:
            # use real restrictedTraverse() here
            obj = self.restrictedTraverse(path, None)
            if obj is None:
                abs_path = os.path.join(module_path, path)
                if not os.path.isfile(abs_path):
                    continue
                fp = open(abs_path)
                data = fp.read()
                fp.close()
                notinzodb.append({ 'path': path, 'meta_type': guess_meta_type(data) })
        return notinzodb

    # standard cvs actions

    security.declareProtected(Permissions.RMManage, "importModule")
    def importModule(self, vendor_tag, release_tag, log_message, paths):
        module = self.getCVSModule()
        # create the import directory
        import_dir = os.tmpnam()
        import_dir = "/tmp/RevisionManager-" + module.filename
        os.mkdir(import_dir)

        # write objects to the file system
        paths = self.getParentAndChildObjectPaths(paths)
        self.writeToFS(paths, import_dir)

        # import the module
        cvs = self._cvs_wrapper
        cvs.importModule(vendor_tag, release_tag, log_message, import_dir=import_dir)        
        # delete the import directory
        deltree(import_dir)

        # do a checkout right away
        self.checkout()
        # now we can say that we're properly initialized, so replace the title
        self.title = self.id
        self.clearLazyTraverse()

    security.declareProtected(Permissions.RMManage, "checkout")
    def checkout(self):
        cvs = self._cvs_wrapper
        cvs.checkout(prune_empty=1)
        if self.haveWorkingCopy():
            # update status information in cvs objects tree
            self.status([''])

    security.declareProtected(Permissions.RMUpdate, "update")
    def update(self, paths):
        if not len(paths):
            return []

        # write objects to FS
        all_paths = self.getParentAndChildFilePaths(paths)
        self.writeToFS(all_paths)

        # do the CVS update
        escaped_paths = self._escape_spaces(paths)
        cvs = self._cvs_wrapper
        cvs_output = cvs.update(" ".join(escaped_paths), prune_empty=1, new_dirs=1)

        # read the files into the objects again
        notinzodb, removed = self.readFromFS(all_paths)

        # prepare the return output
        return_list = []
        updated_paths = []
        for path, status in cvs_output.items():
            extra = ""
            if path in notinzodb:
                extra = "Not in ZODB"
            dict = {'path':   path,
                    'status': status,
                    'extra':  extra }
            return_list.append(dict)
            updated_paths.append(path)

        for path in removed:
            return_list.append({'path':   path,
                                'status': "Removed",
                                'extra':  '' })

        # update the status in CVSObjects
        if len(updated_paths):
            self.status(updated_paths)

        self.clearLazyTraverse()
        return return_list

    security.declareProtected(Permissions.RMCommit, "commit")
    def commit(self, log_message, paths):
        # we assume that this already contains the complete list of modified files
        # so simply write those objects to FS
        self.writeToFS(paths)

        # commit in cvs
        paths = self._escape_spaces(paths)
        cvs = self._cvs_wrapper
        output = cvs.commit(log_message, " ".join(paths))

        # write the FS file contents back again in case it contained $...$
        self.readFromFS(paths)

        self.status(paths)

    security.declareProtected(Permissions.RMAddRemove, "add")
    def add(self, paths):
        settings = self._settings
        all_paths = self.getParentAndChildObjectPaths(paths)
        module_paths = self.getAllFilePaths()
        self.writeToFS(all_paths)

        file_paths = []
        binary_file_paths = []
        dir_paths = []
        for path in all_paths:
            if path in module_paths:
                continue
            
            obj = self.lazyTraverse(path)
            if obj.isPrincipiaFolderish:
                dir_paths.append(path.replace(' ', '\ '))
            else:
                # check for binariness
                filename = self.getAbsoluteFilePath(path)
                fp = open(filename)
                bin = is_binary(fp)
                fp.close()
                
                # escape spaces
                path = path.replace(' ', '\ ')
                if bin:
                    binary_file_paths.append(path)
                else:
                    file_paths.append(path)

        cvs = self._cvs_wrapper
        # first add the directories. They don't need a commit.
        if dir_paths:
            cvs.add(" ".join(dir_paths))
        # now add regular and binary files and commit them.
        file_paths = " ".join(file_paths)
        binary_file_paths = " ".join(binary_file_paths)
        if file_paths:
            cvs.add(file_paths)
        if binary_file_paths:
            cvs.add(binary_file_paths, binary=1)
        cvs.commit("File added to repository by RevisionManager.",
                   file_paths + " " + binary_file_paths)
        # update status
        self.status(paths)
        self.clearLazyTraverse()

    security.declareProtected(Permissions.RMAddRemove, "remove")
    def remove(self, paths):
        # remember all directories in paths. They will be deltree'd
        dir_filenames = []
        for path in paths:
            filename = self.getAbsoluteFilePath(path)
            if os.path.isdir(filename):
                dir_filenames.append(filename)

        # expand the list to include all parent and child paths
        paths = self.getParentAndChildFilePaths(paths)

        file_paths = []
        dir_paths = []
        for path in paths:
            filename = self.getAbsoluteFilePath(path)
            path = path.replace(' ', '\ ')
            if os.path.isdir(filename):
                # remember directories
                dir_paths.append(path)
            elif os.path.isfile(filename):
                # delete regular files right away
                os.unlink(filename)
                file_paths.append(path)

        # remove regular files from cvs
        cvs = self._cvs_wrapper
        if len(file_paths):
            file_paths = " ".join(file_paths)
            cvs.remove(file_paths)
            # file won't be removed from the repository unless they are committed
            cvs.commit("Removed file by RevisionManager.", file_paths)

        # remove directories from cvs
        if len(dir_paths):
            cvs.remove(" ".join(dir_paths))

        # finally, delete directories recursively
        for directory in dir_filenames:
            deltree(directory)
            
        self.clearLazyTraverse()

    security.declareProtected(Permissions.RMManage, "status")
    def status(self, paths):
        if not len(paths):
            return {}
        
        # write objects to FS
        all_paths = self.getParentAndChildFilePaths(paths)
        self.writeToFS(all_paths)

        # update status info from cvs
        paths = self._escape_spaces(paths)
        cvs = self._cvs_wrapper
        status = cvs.status(" ".join(paths))
        module = self.getCVSModule()
        module.setStatusInfo(status)
        self.clearLazyTraverse()
        return status

    security.declareProtected(Permissions.RMManage, "log")
    def log(self, paths):
        if not len(paths):
            return {}

        # write objects to FS
        all_paths = self.getParentAndChildFilePaths(paths)
        self.writeToFS(all_paths)

        # retrieve cvs log information
        paths = self._escape_spaces(paths)
        cvs = self._cvs_wrapper
        self.clearLazyTraverse()
        return cvs.log(" ".join(paths))

    security.declareProtected(Permissions.RMManage, "diff")
    def diff(self, paths, rev1='', rev2=''):
        if not len(paths):
            return ''
        
        # write objects to FS
        all_paths = self.getParentAndChildFilePaths(paths)
        self.writeToFS(all_paths)

        # retrieve cvs diff information
        paths = self._escape_spaces(paths)
        cvs = self._cvs_wrapper
        self.clearLazyTraverse()
        return cvs.diff(" ".join(paths), rev1, rev2)

    security.declareProtected(Permissions.RMManage, "tag")
    def tag(self, tag, paths):
        paths = self._escape_spaces(paths)
        cvs = self._cvs_wrapper
        return cvs.tag(tag, " ".join(paths))

    security.declareProtected(Permissions.RMManage, "merge")
    def merge(self, path, rev1, rev2=""):
        if not rev2:
            rev2 = rev1
            status = self.status([path])
            rev1 = status[path]['wrev']
        escaped_path = path.replace(' ', '\ ')
        cvs = self._cvs_wrapper
        if rev1 == rev2:
            filename = self.getAbsoluteFilePath(path)
            os.unlink(filename)
            cvs.update(escaped_path)
        else:
            cvs.merge(rev1, rev2, escaped_path)
        self.readFromFS([path])
        self.status([path])

    security.declareProtected(Permissions.RMManage, "getModifiedFiles")
    def getLocallyModifiedFiles(self, paths=[]):
        if not len(paths):
            paths = ['']
            
        # write objects to FS
        all_paths = self.getParentAndChildFilePaths(paths)
        self.writeToFS(all_paths)

        # retrieve information about locally modified files
        paths = self._escape_spaces(paths)
        cvs = self._cvs_wrapper
        self.clearLazyTraverse()
        return cvs.getLocallyModifiedFiles(' '.join(paths))

    # file system interaction

    security.declarePrivate("writeToFS")
    def writeToFS(self, paths, module_path=None):
        settings = self._settings
        if module_path is None:
            module_path = os.path.join(settings['work_dir'], self._settings['module'])

        # create module directory if not present already
        if settings['module'] not in os.listdir(settings['work_dir']):
            try:
                os.mkdir(module_path)
            except OSError:
                pass

        error_paths = []
        fake_request = self._create_fake_request()
        for path in paths:
            obj = self.lazyTraverse(path)
            if obj is None:
                error_paths.append(path)
                continue
            abs_path = os.path.join(module_path, path)

            if obj.isPrincipiaFolderish:
                if not os.path.isdir(abs_path):
                    os.mkdir(abs_path)
            else:
                src = self._get_source(obj, fake_request)
                already_fs = os.path.isfile(abs_path)
                # check whether the contents has changed
                if already_fs:
                    fp = open(abs_path, "rb")
                    fs_src = fp.read()
                    fp.close()

                # only write if contents has changed
                if not already_fs or (already_fs and src != fs_src):
                    # open, write to file and close it again
                    fp = open(abs_path, "wb")
                    fp.write(src)
                    fp.close()
        return error_paths

    security.declarePrivate("readFromFS")
    def readFromFS(self, paths):
        settings = self._settings
        notinzodb = []
        removed = []
        module_path = os.path.join(settings['work_dir'], settings['module'])
        fake_request = self._create_fake_request()
        for path in paths:
            obj = self.lazyTraverse(path)
            if obj is None:
                # remember if the corresponding object could not be found
                notinzodb.append(path)
                continue
            
            abs_path = os.path.join(module_path, path)

            # only remember files that have been removed
            if not os.path.exists(abs_path):
                removed.append(path)
                continue

            if not obj.isPrincipiaFolderish:
                fp = open(abs_path, "rb")
                src = fp.read()
                fp.close()
                # only modify if necessary
                if self._get_source(obj, fake_request) != src:
                    self._put_source(obj, src)
        return (notinzodb, removed)

    security.declareProtected(Permissions.RMManage, "haveWorkingCopy")
    def haveWorkingCopy(self):
        if not self._settings.has_key('module'):
            return 0
        settings = self._settings
        module_path = os.path.join(settings['work_dir'], settings['module'])
        if not os.path.isdir(module_path):
            return 0
        cvsdir = os.path.join(module_path, 'CVS')
        if os.path.isdir(cvsdir):
            list = os.listdir(cvsdir)
            if 'Entries' in list and 'Repository' in list and 'Root' in list:
                return 1
        return 0

    security.declareProtected(Permissions.RMManage, 'addFSFileObject')
    def addFSFileObject(self, path, meta_type):
        segments = path.split("/")
        # first make sure that the directory part of the path exists
        # if it doesn't, create as many folders as necessary
        folder = aq_parent(self)
        for id in segments[:-1]:
            subfolder = getattr(folder, id, None)
            if subfolder is None:
                folder.manage_addFolder(id)
                subfolder = getattr(folder, id)
            folder = subfolder

        # now add the object
        # first the the appropriate factory
        factory = FactoryRegistry.getFactory(meta_type)
        id = segments[-1]
        file_name = self.getAbsoluteFilePath(path)
        # read the file contents
        data = open(file_name, "rb").read()
        obj = factory.createObject(id, id, data)
        folder._setObject(id, obj)
        return getattr(folder, id)
    
    #
    ############################################################################
    # Overwritten SimpleItem methods
    #

    def manage_beforeDelete(self, item, container):
        self._clean_up()

    ############################################################################
    # Implementation methods for interface
    # interfaces.RevisionManager.IRevisionManagerZMI

    # management views
    security.declareProtected(Permissions.RMManage,
                              "rm_statusForm", "rm_addObjectsForm", "rm_configForm",
                              "rm_importForm", "rm_mapFSObjectsForm", "rm_detailsForm",
                              "rm_logForm", "rm_diffForm", "rm_commitForm",
                              "rm_updateForm")
    rm_statusForm       = PageTemplateFile("www/status.pt", globals())
    rm_addObjectsForm   = PageTemplateFile("www/addObjects.pt", globals())
    rm_configForm       = PageTemplateFile("www/config.pt", globals())
    rm_importForm       = PageTemplateFile("www/import.pt", globals())
    rm_mapFSObjectsForm = PageTemplateFile("www/mapFSObjects.pt", globals())
    rm_detailsForm      = PageTemplateFile("www/details.pt", globals())
    rm_logForm          = PageTemplateFile("www/log.pt", globals())
    rm_diffForm         = PageTemplateFile("www/diff.pt", globals())
    rm_commitForm       = PageTemplateFile("www/commit.pt", globals())
    rm_updateForm       = PageTemplateFile("www/update.pt", globals())

    security.declareProtected(Permissions.RMAddRemove, "rm_import")
    def rm_import(self, vendor_tag, release_tag, log_message, paths=[], REQUEST=None):
        """
        Properly import selected files
        """
        if not len(paths):
            return MessageDialog(
                title="No selection made!",
                message="You did not select anything to import. Please select one or more files or directories.",
                action="rm_importForm"
                )
        
        module = self.getCVSModule()        
        vendor_tag = vendor_tag or module.filename.upper()
        release_tag = vendor_tag or module.filename.upper() + "1"
        log_message = log_message or "RevisionManager initial import"
        if vendor_tag == release_tag:
            release_tag += "1"

        self.importModule(vendor_tag, release_tag, log_message, paths)

        if REQUEST is not None:
            message = "Module '%s' imported" % module.filename
            return self.rm_statusForm(self, REQUEST, manage_tabs_message=message)

    security.declareProtected(Permissions.RMManage, "rm_updateStatus")
    def rm_updateStatus(self, paths=[], REQUEST=None):
        """
        Update status information.
        """
        if not len(paths):
            return MessageDialog(
                title="No selection made!",
                message="You did not select anything to update its status. Please select one or more files or directories.",
                action="rm_statusForm"
                )
        self.status(paths)
        
        if REQUEST is not None:
            message = "Updated status information"
            return self.rm_statusForm(self, REQUEST, manage_tabs_message=message)

    security.declareProtected(Permissions.RMAddRemove, "rm_mapFSObjects")
    def rm_mapFSObjects(self, items=[], REQUEST=None):
        """
        Map file system files to objects.
        """
        for item in items:
            if item.get('add', 0):
                path = item['path']
                self.addFSFileObject(path, item['meta_type'])
        if REQUEST is not None:
            return self.rm_statusForm(self, REQUEST)

    security.declareProtected(Permissions.RMConfig, "rm_config")
    def rm_config(self, REQUEST):
        """
        Save configuration settings.
        """
        # only take 'module' from REQUEST if settings doesn't already have it
        module = self._settings.get("module", REQUEST.get("module"))
        REQUEST.set("module", module)
        self.saveSettings(REQUEST)
        
        if REQUEST is not None:
            message = "Saved changes"
            return self.rm_configForm(self, REQUEST, manage_tabs_message=message)

    security.declareProtected(Permissions.RMAddRemove, "rm_addObjects")
    def rm_addObjects(self, paths=[], REQUEST=None, RESPONSE=None):
        """
        Add objects to the repository.
        """
        if not len(paths):
            return MessageDialog(
                title="No selection made!",
                message="You did not select anything to add. Please select one or more files or directories.",
                action="rm_addObjectsForm"
                )
        self.add(paths)

        if RESPONSE is not None:
            RESPONSE.redirect(self.absolute_url() + '/rm_statusForm')

    security.declareProtected(Permissions.RMAddRemove, "rm_removeObjects")
    def rm_removeObjects(self, paths=[], REQUEST=None, RESPONSE=None):
        """
        Remove objects from RevisionManager's list and delete their
        equivalents in the CVS repository
        """
        if not len(paths):
            return MessageDialog(
                title="No selection made!",
                message="You did not select anything to remove. Please select one or more files or directories.",
                action="rm_statusForm"
                )
        self.remove(paths)

        if RESPONSE is not None:
            RESPONSE.redirect(self.absolute_url() + '/rm_statusForm')

    security.declareProtected(Permissions.RMCommit, "rm_commit")
    def rm_commit(self, log_message, paths=[], REQUEST=None, RESPONSE=None):
        """
        Commit changes
        """
        if not len(paths):
            return MessageDialog(
                title="No selection made!",
                message="You did not select anything to import. Please select one or more files or directories.",
                action="rm_importForm"
                )
        self.commit(log_message, paths)

        if RESPONSE is not None:
            RESPONSE.redirect(self.absolute_url() + '/rm_statusForm')

    security.declareProtected(Permissions.RMCommit, "rm_tag")
    def rm_tag(self, tag, paths=[], REQUEST=None, RESPONSE=None):
        """
        Set tag on file(s)
        """
        if not len(paths):
            return MessageDialog(
                title="No selection made!",
                message="You did not select anything to tag. Please select one or more files or directories.",
                action="rm_importForm"
                )
        try:
            self.tag(tag, paths)
        except CVSError, e:
            return MessageDialog(title="Bad tag!", message=str(e), action="rm_statusForm")

        if RESPONSE is not None:
            RESPONSE.redirect(self.absolute_url() + '/rm_statusForm')

    security.declareProtected(Permissions.RMUpdate, "rm_merge")
    def rm_merge(self, path, revision, REQUEST=None, RESPONSE=None):
        """
        Overwrite an object's data with its data from an older revision
        """
        self.merge(path, revision)

        if RESPONSE is not None:
            RESPONSE.redirect(self.absolute_url() + '/rm_statusForm')

    #
    ############################################################################
        

#
# object constructors
#

Globals.InitializeClass(RevisionManager)

def manage_addRevisionManager(self, module, REQUEST=None, RESPONSE=None):
    """
    Create an instance of RevisionManager
    """
    # prevent instances in folders whose parents already contain an instance
    # (since it may lead to data inconsistency)
    try:
        rm = aq_acquire(self, "RevisionManager")
    except AttributeError:
        rm = None
    if rm is not None:
        return MessageDialog(
	    title="Detected a RevisionManager instance",
	    message="There already is a RevisionManager instance in this folder or one of its parent folders.",
	    action="manage_workspace"
	   )

    rm = RevisionManager()
    self.Destination()._setObject(rm.id, rm)
    rm.saveSettings(REQUEST.form, module)

    # get the instance acquisition wrapped and fire up second stage
    rm = getattr(self, rm.id)
    rm.goToSecondStage(REQUEST)

manage_addForm = PageTemplateFile("www/add.pt", globals())

#
# functions
#

def guess_meta_type(data):
    """
    Guess the meta type of 'data'
    """
    # This is very simple for now
    # If a token in the table is found, it is assummed that the corresponding
    # meta_type is correct
    
    #XXX missing images for now
    table = {
        '<dtml-':            'DTML Method',
        ' tal:':             'Page Template',
        ' Script (Python) ': 'Script (Python)',
        '<params>':          'Z SQL Method'
        }
    default_meta_type = "File"

    for chunk, meta_type in table.items():
        if data.find(chunk) >= 0:
            return meta_type
    return default_meta_type

def is_binary(file):
    """
    Check whether a file represented by file object 'file' contains binary
    data or text
    """
    chunk_size = 4*1024
    # read the file chunk-wise
    data = file.read(chunk_size)
    while data:
        for char in data:
            # binary files usually contain bytes < ASCII 7
            if char <= '\7':
                file.seek(0)
                return 1
        data = file.read(chunk_size)
    return 0

def deltree(dirname):
    """
    works like 'rm -r <dirname>'
    """
    if not os.path.isdir(dirname):
        return
    for filename in os.listdir(dirname):
        if os.path.isdir(os.path.join(dirname, filename)):
            deltree(os.path.join(dirname, filename))
        else:
            os.unlink(os.path.join(dirname, filename))
    os.rmdir(dirname)
