#
# RevisionManager interfaces
#
# Copyright (c) 2001-2003 by Philipp "philiKON" von Weitershausen
#
# This software is distributed under the terms of the Mozilla Public License
# (MPL)
#

"""
$Id: RevisionManager.py,v 1.2 2003/03/17 13:49:52 philipp Exp $
"""

from Interface import Interface, Attribute

class IRevisionManager(Interface):
    """
    Pure python interface for RevisionManager
    """

    # settings

    def getSettings():
        """
        Return dictionary containing settings.
        """

    def setSettings(settings):
        """
        Set settings according to the mapping object parameter 'settings'.
        """

    def saveSettings(mapping):
        """
        Save the settings described in the mapping object parameter 'mapping'.
        
        Optionally:
        - Clean up things that result from changing the settings.
        - Pass the information to delegated class(es).
        - Prepare further things that result from changing the settings.
        """

    # lazy traverse

    def lazyTraverse(path):
        """
        Lazy traverse for 'path'.
        """

    def setLazyTraverse(path, obj):
        """
        Set a lazy traverse shortcut to 'obj' identified by 'path'.
        """

    def clearLazyTraverse():
        """
        Clear lazy traversal shortcut table.
        """

    # 'getters'

    def getModuleFolder():
        """
        Return the Zope folder which is the root of the module.
        """

    def getTreeWrappedModuleFolder():
        """
        Return the module folder wrapped with TreeObjectWrapper.
        """

    def getCVSModule():
        """
        Return module object.
        """

    def traverseCVSObject(path):
        """
        Return a cvs object representing the file or directory identified by 'path'.
        """

    def getAbsoluteFilePath(path):
        """
        Return the absolute file path for the file identified by the relative file
        path 'path'.
        """

    def getSupportedFactoryMetaTypes():
        """
        Return a sequence with all meta types supported by the factory
        """

    def getSubobjectsPaths(path, obj=None):
        """
        Return a list of relative path names of all subobjects of the object
        identified by 'path'. The object identified by 'path' can optionally
        be provided with the 'obj' parameter
        """

    def getParentPaths(path):
        """
        Return a sequence of path names containing all path names in 'paths' and
        the path names leading to them.
        """

    def getAllFilePaths():
        """
        Return the relative paths of all files in the module.
        """

    def getParentAndChildObjectPaths(paths):
        """
        Expands the sequence 'paths' which is supposed to contain paths to
        objects with all parent and child paths.
        """

    def getParentAndChildFilePaths(paths):
        """
        Works like 'getParentAndChildObjectPaths()' but on the file system.
        """

    def getFSFilesNotInZODB():
        """
        Return a list of paths pointing to file system files that are not in the
        ZODB.
        """

    # standard cvs actions

    def importModule(vendor_tag, release_tag, log_message, paths):
        """
        Import the module.

        Use objects specified in 'paths' as initial module content.
        """

    def checkout():
        """
        Checkout the module from repository.
        """

    def update(paths):
        """
        Update objects from repository.
        """

    def commit(log_message, paths):
        """
        Commit objects.
        """

    def add(paths):
        """
        Add objects specified in 'paths' to the module
        """

    def remove(paths):
        """
        Remove objects specified in 'paths' from the module
        """

    def status(paths):
        """
        Return the status.
        """

    def log(paths):
        """
        List items and their information in the cvs log.
        """

    def diff(paths, rev1="", rev2=""):
        """
        Compare two revisions of an object identified by 'path' and return the diff
        as string.
        """

    def tag(tag, paths):
        """
        Tag objects specified in 'paths' with 'tag'.
        """

    def merge(path, rev1, rev2=''):
        """
        Merge differences between 'rev1' and 'rev2' into 'path'.
        """

    def getLocallyModifiedFiles(paths=[]):
        """
        Return a list of locally modified files identified by their path
        """

    # file system interaction

    def writeToFS(paths, module_path=None):
        """
        Write objects identified by 'paths' to the filesystem
        """

    def readFromFS(paths):
        """
        Read data for those objects identified by 'paths' from the filesystem
        and upload them into the objects.
        """

    def haveWorkingCopy():
        """
        Return true if there is a properly checked out working copy of the current
        module available.
        """

    def addFSFileObject(path, meta_type):
        """
        Add file system file identified by 'path' as an object of type 'meta_type'
        """

class IRevisionManagerZMI(Interface):
    """
    This interface describes methods and views for RevisionManager's ZMI
    interface. Use methods described in IRevisionManager from pure python.
    """
    
    rm_importForm = Attribute("""
        Import form. Show objects that can be imported.
        """)
        
    rm_mapFSObjectsForm = Attribute("""
        Mapping file system files to objects form. Show a list of appropriate files.
        """)

    rm_configForm = Attribute("""
        Configuration form.
        """)
        
    rm_addObjectsForm = Attribute("""
        Add objects form.
        """)

    rm_commitForm = Attribute("""
        Commit form.
        """)
        
    rm_statusForm = Attribute("""
        Status form. Show objects' status.
        """)

    rm_updateForm = Attribute("""
        Update form. Display update output.
        """)

    rm_detailsForm = Attribute("""
        Details form. Show detailled status information about objects.
        """)

    rm_logForm = Attribute("""
        log form. Show an object's history log.
        """)

    rm_diffForm = Attribute("""
        Diff form. Show the diff of an object's two revisions.
        """)

    def rm_import(vendor_tag, release_tag, log_message, paths=[], REQUEST=None):
        """
        Import files specified in the 'paths' sequence.
        """

    def rm_updateStatus(paths=[], REQUEST=None):
        """
        Update status information.
        """

    def rm_mapFSObjects(items=[], REQUEST=None):
        """
        Map file system files to objects. Parameter 'items' is a sequence of dictionaries.
        """

    def rm_config(REQUEST):
        """
        Save configuration settings.
        """

    def rm_addObjects(paths=[], REQUEST=None, RESPONSE=None):
        """
        Add objects to the repository.
        """

    def rm_removeObjects(paths=[], REQUEST=None, RESPONSE=None):
        """
        remove objects from RevisionManager's list and delete their
        equivalents in the CVS repository.
        """

    def rm_commit(paths=[], log_message="commit", REQUEST=None):
        """
        Commit changes.
        """

    def rm_tag(tag, paths=[], REQUEST=None):
        """
        Set tag 'tag' on files specified in 'paths'.
        """

    def rm_merge(path, revision, REQUEST=None, RESPONSE=None):
        """
        Merge the differences between HEAD and 'revision' into 'path'.
        """
